import contextlib
import json
from datetime import datetime, timezone
from zoneinfo import ZoneInfo

from django.db.utils import IntegrityError
from taggit.models import Tag

from ditto.flickr.models import Photo, Photoset, User

from . import FetchError

# These classes are passed JSON data from the Flickr API and create/update
# objects based on that.
#
# They don't do any fetching of data or files themselves.

# CLASSES HERE:
#
# SaveUtilsMixin
#
# UserSaver
# PhotoSaver
# PhotosetSave


class SaveUtilsMixin:
    """Handy utility methods used by the *Saver classes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _api_datetime_to_datetime(self, api_time, timezone_id=None):
        """Change a text datetime from the API to a datetime with timezone.
        api_time is a string like "1956-01-01 00:00:00".
        timezone_id is a TZ timezone name like 'Africa/Accra'
        """
        if timezone_id is None or timezone_id == "":
            timezone_id = "Etc/UTC"
        tz = ZoneInfo(timezone_id)
        return datetime.strptime(api_time, "%Y-%m-%d %H:%M:%S").replace(tzinfo=tz)

    def _unixtime_to_datetime(self, api_time):
        """Change a text unixtime from the API to a datetime with timezone.
        api_time is a string or int like "1093459273".
        """
        return datetime.fromtimestamp(int(api_time), tz=timezone.utc)


class UserSaver(SaveUtilsMixin):
    """For creating/updating an individual User based on data from the API.

    Use like:

        UserSaver().save_user(data, fetch_time)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def save_user(self, user, fetch_time):
        """
        Keyword arguments:
        user -- A dict of the data about a user from the API's JSON.
        fetch_time -- A datetime.

        Returns the User object that was created or updated.
        """
        raw_json = json.dumps(user)

        defaults = {
            "fetch_time": fetch_time,
            "raw": raw_json,
            "nsid": user["nsid"],
            "username": user["username"]["_content"],
            "photos_count": int(user["photos"]["count"]["_content"]),
            "photos_first_date": self._unixtime_to_datetime(
                user["photos"]["firstdate"]["_content"]
            ),
            "photos_first_date_taken": self._api_datetime_to_datetime(
                user["photos"]["firstdatetaken"]["_content"]
            ),
        }

        # Things we won't have in our dummy data if the user was deleted:
        if "iconserver" in user:
            defaults["iconserver"] = user["iconserver"]
        if "iconfarm" in user:
            defaults["iconfarm"] = user["iconfarm"]
        if "ispro" in user:
            defaults["is_pro"] = int(user["ispro"]) == 1
        if "description" in user:
            defaults["description"] = user["description"]["_content"]
        if "photosurl" in user:
            defaults["photos_url"] = user["photosurl"]["_content"]
        if "profileurl" in user:
            defaults["profile_url"] = user["profileurl"]["_content"]

        # Surprisingly, not always present:
        if "realname" in user:
            defaults["realname"] = user["realname"]["_content"]

        # 'location' and 'timezone' were missing for at least one user fetched:
        if "location" in user:
            defaults["location"] = user["location"]["_content"]
        if "timezone" in user:
            defaults["timezone_id"] = user["timezone"]["timezone_id"]

        if "views" in user["photos"]:
            # I think this might only be returned for the Account's own user.
            defaults["photos_views"] = int(user["photos"]["views"]["_content"])

        user_obj, created = User.objects.update_or_create(
            nsid=user["nsid"], defaults=defaults
        )

        return user_obj


class PhotoSaver(SaveUtilsMixin):
    """For creating/updating an individual Photo based on data from the API.

    Use like:

        PhotoSaver().save_photo(data, fetch_time)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def save_photo(self, photo):
        """Takes a dict of photo data from the API and creates or updates a
        Photo object.

        Keyword arguments:
        photo -- The photo data, from several Flickr API calls.
                 Has keys:
                     'fetch_time': A datetime.
                     'user_obj': User object for this photo's owner.
                     'info': Results of a photos.getInfo call.
                     'exif': Results of a photos.getExif call.
                     'sizes': Results of a photos.getSizes call.

        Returns:
        The Photo object that was created or updated.
        """

        # photo['info']['urls'] = {
        #     'url': [ {'type':'photopage', '_content':'http...'} ]
        # }
        permalink = next(
            url for url in photo["info"]["urls"]["url"] if url["type"] == "photopage"
        )["_content"]

        defaults = {
            # DittoItemModel fields
            "title": photo["info"]["title"]["_content"],
            "permalink": permalink,
            "is_private": (photo["info"]["visibility"]["ispublic"] == 0),
            "fetch_time": photo["fetch_time"],
            "post_time": self._unixtime_to_datetime(photo["info"]["dates"]["posted"]),
            "raw": json.dumps(photo["info"]),
            # Photo fields
            "user": photo["user_obj"],
            "flickr_id": photo["info"]["id"],
            "description": photo["info"]["description"]["_content"],
            "secret": photo["info"]["secret"],
            "original_secret": photo["info"]["originalsecret"],
            "server": photo["info"]["server"],
            "farm": photo["info"]["farm"],
            "license": photo["info"]["license"],
            "rotation": photo["info"]["rotation"],
            "original_format": photo["info"]["originalformat"],
            "safety_level": photo["info"]["safety_level"],
            "has_people": (int(photo["info"]["people"]["haspeople"]) == 1),
            "last_update_time": self._unixtime_to_datetime(
                photo["info"]["dates"]["lastupdate"]
            ),
            "taken_time": self._api_datetime_to_datetime(
                photo["info"]["dates"]["taken"], photo["user_obj"].timezone_id
            ),
            "taken_granularity": int(photo["info"]["dates"]["takengranularity"]),
            "taken_unknown": (int(photo["info"]["dates"]["takenunknown"]) == 1),
            "view_count": photo["info"]["views"],
            "comment_count": photo["info"]["comments"]["_content"],
            "media": photo["info"]["media"],
            # Location fields
            # Sizes fields
            "sizes_raw": json.dumps(photo["sizes"]),
            # EXIF fields
            "exif_raw": json.dumps(photo["exif"]),
            "exif_camera": photo["exif"]["camera"],
        }

        # Photos with no location have these fields missing entirely:
        if "geoperms" in photo["info"]:
            defaults["geo_is_private"] = int(photo["info"]["geoperms"]["ispublic"]) == 0

        if "location" in photo["info"]:
            loc = photo["info"]["location"]
            defaults["latitude"] = loc["latitude"]
            defaults["longitude"] = loc["longitude"]
            defaults["location_accuracy"] = loc["accuracy"]
            defaults["location_context"] = loc["context"]

            if "place_id" in loc:
                defaults["location_place_id"] = loc["place_id"]
            if "woeid" in loc:
                defaults["location_woeid"] = loc["woeid"]

            if "locality" in loc:
                defaults["locality_name"] = loc["locality"]["_content"]
                if "place_id" in loc["locality"]:
                    defaults["locality_place_id"] = loc["locality"]["place_id"]
                if "woeid" in loc["locality"]:
                    defaults["locality_woeid"] = loc["locality"]["woeid"]

            if "county" in loc:
                defaults["county_name"] = loc["county"]["_content"]
                if "place_id" in loc["county"]:
                    defaults["county_place_id"] = loc["county"]["place_id"]
                if "woeid" in loc["county"]:
                    defaults["county_woeid"] = loc["county"]["woeid"]

            if "region" in loc:
                defaults["region_name"] = loc["region"]["_content"]
                if "place_id" in loc["region"]:
                    defaults["region_place_id"] = loc["region"]["place_id"]
                if "woeid" in loc["region"]:
                    defaults["region_woeid"] = loc["region"]["woeid"]

            if "country" in loc:
                defaults["country_name"] = loc["country"]["_content"]
                if "place_id" in loc["country"]:
                    defaults["country_place_id"] = loc["country"]["place_id"]
                if "woeid" in loc["country"]:
                    defaults["country_woeid"] = loc["country"]["woeid"]

        # The size labels for all possible sizes an image might have, that we
        # also have width/height parameters for on Photo:
        sizes = [v["label"] for k, v in Photo.PHOTO_SIZES.items()] + [
            v["label"] for k, v in Photo.VIDEO_SIZES.items()
        ]
        # We don't store width/height for these, so ignore them:
        sizes.remove("Square")
        sizes.remove("Large square")

        for size in photo["sizes"]["size"]:
            if size["label"] in sizes:
                # eg, 'X-Large 3K' becomes 'x_large_3k':
                name = size["label"].lower().replace(" ", "_").replace("-", "_")
                if size["width"] is not None:
                    defaults[name + "_width"] = int(size["width"])
                if size["height"] is not None:
                    defaults[name + "_height"] = int(size["height"])

        try:
            for e in photo["exif"]["exif"]:
                if e["tag"] == "LensModel":
                    defaults["exif_lens_model"] = e["raw"]["_content"]
                elif e["tag"] == "FNumber":
                    defaults["exif_aperture"] = e["clean"]["_content"]
                elif e["tag"] == "ExposureTime":
                    defaults["exif_exposure"] = e["clean"]["_content"]
                elif e["tag"] == "Flash":
                    defaults["exif_flash"] = e["raw"]["_content"]
                elif e["tag"] == "FocalLength":
                    defaults["exif_focal_length"] = e["clean"]["_content"]
                elif e["tag"] == "ISO":
                    defaults["exif_iso"] = e["raw"]["_content"]
        except KeyError:
            pass

        photo_obj, created = Photo.objects.update_or_create(
            flickr_id=photo["info"]["id"], defaults=defaults
        )

        self._save_tags(photo_obj, photo["info"]["tags"]["tag"])

        return photo_obj

    def _save_tags(self, photo_obj, tags_data):
        """
        Adds/deletes tags for a photo.

        Required Arguments
          photo_obj: The Photo object we're altering tags for.
          photo_user_obj: The User whose photo this is (because they're most
                            likely to have created the tags.)
          tags_data: A list of dicts about the tags, straight from the API.
        """

        # The existing tag-photo relationships.
        tagged_photos = Photo.tags.through.objects.filter(content_object=photo_obj)

        local_flickr_ids = set()
        remote_flickr_ids = set()

        # Get the Flickr IDs of all the current tag-photo relationships.
        for tagged_photo in tagged_photos:
            local_flickr_ids.add(tagged_photo.flickr_id)

        for tag in tags_data:
            remote_flickr_ids.add(tag["id"])

            if tag["id"] not in local_flickr_ids:
                # This tag isn't currently on the photo, so add it.
                try:
                    tag_obj, tag_created = Tag.objects.get_or_create(
                        slug=tag["_content"], defaults={"name": tag["raw"]}
                    )
                except IntegrityError:
                    # It's possible for there to be a tag with a different
                    # slug but the same name, which would cause an
                    # IntegrityError.
                    # In which case, just fetch the existing Tag by name:
                    tag_obj = Tag.objects.get(name=tag["raw"])

                # Who created this tag?
                if tag["author"] == photo_obj.user.nsid:
                    # The same person whose photo these tags are on.
                    user = photo_obj.user
                else:
                    # In theory we'll already have fetched and saved data for
                    # all authors of these tags when fetching this photo's
                    # data.
                    try:
                        user = User.objects.get(nsid=tag["author"])
                    except User.DoesNotExist as err:
                        msg = (
                            "Tried to add a Tag authored by a Flickr user "
                            f"with NSID {tag['author']} who doesn't exist in the DB."
                        )
                        raise FetchError(msg) from err

                pt_obj = Photo.tags.through(
                    flickr_id=tag["id"],
                    author=user,
                    machine_tag=(tag["machine_tag"] == "1"),
                    content_object=photo_obj,
                    tag=tag_obj,
                )
                pt_obj.save()

        flickr_ids_to_delete = local_flickr_ids.difference(remote_flickr_ids)

        # Finally, delete any tag-photo relationships which were identified
        # above as no longer on the photo on Flickr.
        for tagged_photo in tagged_photos:
            if tagged_photo.flickr_id in flickr_ids_to_delete:
                tagged_photo.delete()


class PhotosetSaver(SaveUtilsMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def save_photoset(self, photoset):
        """Takes a dict of photoset data from the API and creates or updates a
        Photoset object.

        It accepts a list of photo data, but will only add each photo to the
        photoset if we already have that photo's data in the DB.

        Keyword arguments:
        photoset -- Has keys:
                    'fetch_time': A datetime.
                    'photoset': The data about the photoset from the API.
                    'photos': The data about the set's photos from the API.
                    'user_obj': User object for this photoset's owner.

        Returns:
        The Photoset object that was created or updated.
        """

        ps = photoset["photoset"]

        defaults = {
            "fetch_time": photoset["fetch_time"],
            "user": photoset["user_obj"],
            "flickr_id": ps["id"],
            "title": ps["title"]["_content"],
            "description": ps["description"]["_content"],
            "photo_count": ps["photos"],
            "video_count": ps["videos"],
            "view_count": ps["count_views"],
            "comment_count": ps["count_comments"],
            "last_update_time": self._unixtime_to_datetime(ps["date_update"]),
            "flickr_created_time": self._unixtime_to_datetime(ps["date_create"]),
            "raw": json.dumps(ps),
            "photos_raw": json.dumps(photoset["photos"]),
        }

        with contextlib.suppress(Photo.DoesNotExist):
            defaults["primary_photo"] = Photo.objects.get(flickr_id=ps["primary"])

        photoset_obj, created = Photoset.objects.update_or_create(
            flickr_id=ps["id"], defaults=defaults
        )

        if photoset_obj:
            # Add all the photoset's photos that we have in the DB to the
            # photoset object.
            photos = []
            for photo in photoset["photos"]:
                with contextlib.suppress(Photo.DoesNotExist):
                    photos.append(Photo.objects.get(flickr_id=photo["id"]))

            # Sets/updates the SortedManyToMany field of the photoset's photos:
            photoset_obj.photos.set(photos)

        return photoset_obj



from datetime import datetime, timezone

from django.db.models import Count
from django.utils.html import strip_tags
from django.utils.text import Truncator


def truncate_string(
    text, *, strip_html=True, chars=255, truncate="…", at_word_boundary=False
):
    """Truncate a string to a certain length, removing line breaks and mutliple
    spaces, optionally removing HTML, and appending a 'truncate' string.

    Keyword arguments:
    strip_html -- boolean.
    chars -- Number of characters to return.
    at_word_boundary -- Only truncate at a word boundary, which will probably
        result in a string shorter than chars.
    truncate -- String to add to the end.
    """
    if strip_html:
        text = strip_tags(text)
    text = text.replace("\n", " ").replace("\r", "")
    text = " ".join(text.split())
    if at_word_boundary:
        if len(text) > chars:
            text = text[:chars].rsplit(" ", 1)[0] + truncate
    else:
        text = Truncator(text).chars(chars, html=False, truncate=truncate)
    return text


def datetime_now():
    """Just returns a datetime object for now in UTC, with UTC timezone.
    Because I was doing this a lot in various places.
    """
    return datetime.now(tz=timezone.utc)


def datetime_from_str(s):
    """A shortcut for making a UTC datetime from a string like
    '2015-08-11 12:00:00'.
    """
    return datetime.strptime(s, "%Y-%m-%d %H:%M:%S").replace(tzinfo=timezone.utc)


def get_annual_item_counts(qs, field_name="post_year"):
    """
    Takes a QuerySet, probably of a DittoItem child class like Photo or Tweet,
    and returns a list of dicts with 'year' and 'count' keys. eg:
        [
            {'year': 2015, 'count': 123},
            {'year': 2016, 'count': 456},
            {'year': 2017, 'count': 0},
            {'year': 2018, 'count': 789},
        ]

    Note, the first query we do here gets a QuerySet that only includes years
    in which there are any counts:
        [
            {'post_year': 2015, 'count': 123},
            {'post_year': 2016, 'count': 456},
            # Nothing included for 2017, which has count 0.
            {'post_year': 2018, 'count': 789},
        ]

    So the rest of this function is all about filling in those gaps.

    Used in all the annual_[thing]_counts() template tags.

    Arguments:
        qs -- The QuerySet.
        field_name -- Name of the key identifying the years.
    """

    qs = (
        qs.values(field_name)
        .annotate(count=Count("id"))
        .values(field_name, "count")
        .order_by(field_name)
    )

    # Just in case. eg, trying to get counts for a private Twitter account:
    if len(qs) == 0:
        return []

    # Make a set of years like {2015, 2016, 2018}:
    years_with_counts = {y[field_name] for y in qs}

    # Make a set of years with no gaps like {2015, 2016, 2017, 2018}:
    all_years = sorted(set(range(min(years_with_counts), max(years_with_counts) + 1)))

    # Translate original qs into {'2015': 123, '2016': 456, '2018': 789}:
    qs_dict = {}
    for row in qs:
        qs_dict[str(row[field_name])] = row["count"]

    # Make the final list of dicts.
    # An element for every year between min and max, even if it had no items,
    # in which case its count is 0.
    results = []
    for y in all_years:
        try:
            count = qs_dict[str(y)]
        except KeyError:
            count = 0
        results.append({"year": y, "count": count})

    return results



import json

from django.db import migrations


def backfill_photo_sizes(apps, schema_editor):
    """
    Go through all Photos and look at its raw Sizes JSON data, looking
    at the larger sizes that didn't originally exist.
    If a Photo has one of those sizes set in the JSON, but not set in
    its model fields, then set the width/height fields.

    This would have happened if the sizes existed in the Flickr API
    when we last fetched the Photo's data, but on that date we didn't
    yet have the corresponding model fields.

    This won't fetch new JSON data, so sizes added to the API since
    the data was last fetched for the photo won't be updated on the
    model fields.
    """
    Photo = apps.get_model("flickr", "Photo")

    # The sizes we'll backfill, if they exist in the JSON data we
    # already have.
    size_fields = (
        ("Medium 800", "medium_800"),
        ("Large", "large"),
        ("Large 1600", "large_1600"),
        ("Large 2048", "large_2048"),
        ("X-Large 3K", "x_large_3k"),
        ("X-Large 4K", "x_large_4k"),
        ("X-Large 5K", "x_large_5k"),
        ("X-Large 6K", "x_large_6k"),
    )

    for photo in Photo.objects.all():
        sizes_raw = json.loads(photo.sizes_raw)
        for size in sizes_raw["size"]:
            # size has keys "label", "width", "height"...
            for field in size_fields:
                width_field = "%s_width" % field[1]
                height_field = "%s_height" % field[1]
                if field[0] == size["label"] and getattr(photo, width_field) is None:
                    setattr(photo, width_field, size["width"])
                    setattr(photo, height_field, size["height"])
                    photo.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flickr", "0024_auto_20201119_1539"),
    ]

    operations = [migrations.RunPython(backfill_photo_sizes)]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("twitter", "0015_auto_20150819_1349"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="account",
            options={"ordering": ["user__screen_name"]},
        ),
        migrations.AlterField(
            model_name="account",
            name="consumer_key",
            field=models.CharField(help_text="(API Key)", blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name="account",
            name="consumer_secret",
            field=models.CharField(
                help_text="(API Secret)", blank=True, max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="last_favorite_id",
            field=models.BigIntegerField(
                help_text="The Twitter ID of the most recent favorited Tweet fetched.",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="last_recent_id",
            field=models.BigIntegerField(
                help_text="The Twitter ID of the most recent Tweet fetched.",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="time_created",
            field=models.DateTimeField(
                help_text="The time this item was created in the database.",
                auto_now_add=True,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="time_modified",
            field=models.DateTimeField(
                help_text="The time this item was last saved to the database.",
                auto_now=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="created_at",
            field=models.DateTimeField(
                help_text="UTC time when this Tweet was created on Twitter"
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="favorite_count",
            field=models.PositiveIntegerField(
                help_text="Approximately how many times this had been favorited when fetched",  # noqa: E501
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="fetch_time",
            field=models.DateTimeField(
                help_text="The time the item's data was last fetched, and was new or changed.",  # noqa: E501
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="in_reply_to_screen_name",
            field=models.CharField(
                help_text="Screen name of the original Tweet's author, if this is a reply",  # noqa: E501
                blank=True,
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="in_reply_to_status_id",
            field=models.BigIntegerField(
                help_text="The ID of the Tweet replied to, if any",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="in_reply_to_user_id",
            field=models.BigIntegerField(
                help_text="ID of the original Tweet's author, if this is a reply",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="is_private",
            field=models.BooleanField(
                help_text="If set, this item will not be shown on public-facing pages.",
                default=False,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="language",
            field=models.CharField(
                help_text="A BCP 47 language identifier, or 'und' if it couldn't be detected",  # noqa: E501
                default="und",
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="permalink",
            field=models.URLField(
                help_text="URL of the item on the service's website.", blank=True
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="quoted_status_id",
            field=models.BigIntegerField(
                help_text="The ID of the Tweet quoted, if any", null=True, blank=True
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="raw",
            field=models.TextField(
                help_text="eg, the raw JSON from the API.", blank=True
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="retweet_count",
            field=models.PositiveIntegerField(
                help_text="Number of times this had been retweeted when fetched",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="source",
            field=models.CharField(
                help_text="Utility used to post the Tweet", blank=True, max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="summary",
            field=models.CharField(
                help_text="eg, Initial text of a blog post, start of the description of a photo, all of a Tweet's text, etc. No HTML.",  # noqa: E501
                blank=True,
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="time_created",
            field=models.DateTimeField(
                help_text="The time this item was created in the database.",
                auto_now_add=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="time_modified",
            field=models.DateTimeField(
                help_text="The time this item was last saved to the database.",
                auto_now=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="created_at",
            field=models.DateTimeField(
                help_text="UTC time when this account was created on Twitter"
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="favorites_count",
            field=models.PositiveIntegerField(
                help_text="The number of tweets this user has favorited in the account’s lifetime",  # noqa: E501
                default=0,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="fetch_time",
            field=models.DateTimeField(
                help_text="The time the data was last fetched, and was new or changed.",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="followers_count",
            field=models.PositiveIntegerField(
                help_text="The number of followers this account has", default=0
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="friends_count",
            field=models.PositiveIntegerField(
                help_text="Tne number of users this account is following.", default=0
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="is_private",
            field=models.BooleanField(
                help_text="True if this user is 'protected'", default=False
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="listed_count",
            field=models.PositiveIntegerField(
                help_text="The number of public lists this user is a member of",
                default=0,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="name",
            field=models.CharField(help_text="eg, 'Samuel Pepys'", max_length=30),
        ),
        migrations.AlterField(
            model_name="user",
            name="raw",
            field=models.TextField(
                help_text="eg, the raw JSON from the API.", blank=True
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="screen_name",
            field=models.CharField(
                help_text="Username, eg, 'samuelpepys'", max_length=20
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="statuses_count",
            field=models.PositiveIntegerField(
                help_text="The number of tweets, including retweets, by this user",
                default=0,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="time_created",
            field=models.DateTimeField(
                help_text="The time this item was created in the database.",
                auto_now_add=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="time_modified",
            field=models.DateTimeField(
                help_text="The time this item was last saved to the database.",
                auto_now=True,
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="url",
            field=models.URLField(
                help_text="A URL provided by the user as part of their profile",
                blank=True,
                default="",
            ),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-23 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flickr", "0026_alter_taggedphoto_content_object_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="photos_url",
            field=models.URLField(
                blank=True, max_length=255, verbose_name="Photos URL at Flickr"
            ),
        ),
        migrations.AlterField(
            model_name="user",
            name="profile_url",
            field=models.URLField(
                blank=True, max_length=255, verbose_name="Avatar URL on Flickr"
            ),
        ),
    ]



from django.contrib import admin

from ditto.core.admin import DittoItemModelAdmin

from .models import Account, Album, Artist, Scrobble, Track


@admin.register(Account)
class AccountAdmin(admin.ModelAdmin):
    list_display = (
        "username",
        "realname",
        "is_active",
        "time_created",
    )
    search_fields = (
        "username",
        "realname",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "username",
                    "realname",
                    "api_key",
                    "is_active",
                )
            },
        ),
        (
            "Data",
            {
                "fields": (
                    "time_created",
                    "time_modified",
                )
            },
        ),
    )

    readonly_fields = (
        "time_created",
        "time_modified",
    )


@admin.register(Artist)
class ArtistAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "mbid",
        "time_created",
    )
    search_fields = (
        "name",
        "mbid",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "slug",
                    "original_slug",
                    "mbid",
                )
            },
        ),
        (
            "Data",
            {
                "fields": (
                    "time_created",
                    "time_modified",
                )
            },
        ),
    )

    readonly_fields = (
        "time_created",
        "time_modified",
    )


@admin.register(Track)
class TrackAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "artist",
        "mbid",
        "time_created",
    )
    search_fields = (
        "name",
        "artist__name",
        "mbid",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "slug",
                    "original_slug",
                    "artist",
                    "mbid",
                )
            },
        ),
        (
            "Data",
            {
                "fields": (
                    "time_created",
                    "time_modified",
                )
            },
        ),
    )

    readonly_fields = (
        "time_created",
        "time_modified",
    )
    raw_id_fields = ("artist",)


@admin.register(Album)
class AlbumAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "artist",
        "mbid",
        "time_created",
    )
    search_fields = (
        "name",
        "artist__name",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "slug",
                    "original_slug",
                    "artist",
                    "mbid",
                )
            },
        ),
        (
            "Data",
            {
                "fields": (
                    "time_created",
                    "time_modified",
                )
            },
        ),
    )

    readonly_fields = (
        "time_created",
        "time_modified",
    )
    raw_id_fields = ("artist",)


@admin.register(Scrobble)
class ScrobbleAdmin(DittoItemModelAdmin):
    list_display = (
        "title",
        "post_time",
    )
    list_filter = ("post_time",)
    search_fields = (
        "title",
        "track__name",
    )

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "account",
                    "artist",
                    "track",
                    "album",
                    "post_time",
                    "post_year_str",
                )
            },
        ),
        (
            "Data",
            {
                "fields": (
                    "raw",
                    "fetch_time",
                    "time_created",
                    "time_modified",
                )
            },
        ),
    )

    raw_id_fields = (
        "artist",
        "track",
        "album",
    )
    readonly_fields = (
        "post_year_str",
        "raw",
        "fetch_time",
        "time_created",
        "time_modified",
    )



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("twitter", "0018_tweet_text_html"),
    ]

    operations = [
        migrations.CreateModel(
            name="Photo",
            fields=[
                (
                    "id",
                    models.AutoField(
                        primary_key=True,
                        auto_created=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time this item was created in the database.",
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time this item was last saved to the database.",
                    ),
                ),
                ("twitter_id", models.BigIntegerField(unique=True)),
                ("media_url", models.URLField(help_text="URL of the image itself")),
                (
                    "page_url",
                    models.URLField(help_text="URL of its page on Twitter.com"),
                ),
                (
                    "is_private",
                    models.BooleanField(
                        default=False,
                        help_text="If true, this item will not be shown on public-facing pages.",  # noqa: E501
                    ),
                ),
                (
                    "large_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Large width", blank=True
                    ),
                ),
                (
                    "large_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Large height", blank=True
                    ),
                ),
                (
                    "medium_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Medium width", blank=True
                    ),
                ),
                (
                    "medium_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Medium height", blank=True
                    ),
                ),
                (
                    "small_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Small width", blank=True
                    ),
                ),
                (
                    "small_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Small height", blank=True
                    ),
                ),
                (
                    "thumb_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Thumbnail width", blank=True
                    ),
                ),
                (
                    "thumb_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Thumbnail height", blank=True
                    ),
                ),
            ],
            options={"ordering": ["time_created"]},
        ),
        migrations.AddField(
            model_name="tweet",
            name="photos_count",
            field=models.PositiveSmallIntegerField(
                default=0,
                help_text="Number of Photos attached to this Tweet",
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="tweet",
            name="is_private",
            field=models.BooleanField(
                default=False,
                help_text="If true, this item will not be shown on public-facing pages.",  # noqa: E501
            ),
        ),
        migrations.AddField(
            model_name="photo",
            name="tweet",
            field=models.ForeignKey(to="twitter.Tweet", on_delete=models.CASCADE),
        ),
    ]



# Generated by Django 1.9.5 on 2016-04-29 18:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flickr", "0009_photoset_fetch_time"),
    ]

    operations = [
        migrations.AddField(
            model_name="photo",
            name="photosets",
            field=models.ManyToManyField(to="flickr.Photoset"),
        ),
    ]



from django.conf import settings
from django.core.checks import Error, Tags, register


@register(Tags.compatibility)
def check_taggit_is_installed(app_configs=None, **kwargs):
    checks = []
    try:
        from taggit import models  # noqa: F401

    except ImportError:
        checks.append(
            Error(
                "The django-taggit app is required to use ditto.pinboard.",
                hint=("Install django-taggit"),
                id="ditto.pinboard.E001",
            )
        )

    if len(checks) == 0 and "taggit" not in settings.INSTALLED_APPS:
        checks.append(
            Error(
                "The django-taggit app must be in INSTALLED_APPS",
                hint=("Add 'taggit' to INSTALLED_APPS in your settings file."),
                id="ditto.pinboard.E002",
            )
        )

    return checks



import taggit.managers
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("pinboard", "0012_account_is_active"),
    ]

    operations = [
        migrations.CreateModel(
            name="BookmarkTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        auto_created=True,
                        verbose_name="ID",
                        primary_key=True,
                    ),
                ),
                (
                    "name",
                    models.CharField(unique=True, verbose_name="Name", max_length=100),
                ),
                (
                    "slug",
                    models.SlugField(unique=True, verbose_name="Slug", max_length=100),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        help_text="The time this item was created in the database.",
                        auto_now_add=True,
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        help_text="The time this item was last saved to the database.",
                        auto_now=True,
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="TaggedBookmark",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        auto_created=True,
                        verbose_name="ID",
                        primary_key=True,
                    ),
                ),
                (
                    "object_id",
                    models.IntegerField(verbose_name="Object id", db_index=True),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        help_text="The time this item was created in the database.",
                        auto_now_add=True,
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        help_text="The time this item was last saved to the database.",
                        auto_now=True,
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        related_name="pinboard_taggedbookmark_tagged_items",
                        to="contenttypes.ContentType",
                        verbose_name="Content type",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "tag",
                    models.ForeignKey(
                        related_name="pinboard_taggedbookmark_items",
                        to="pinboard.BookmarkTag",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AlterField(
            model_name="bookmark",
            name="tags",
            field=taggit.managers.TaggableManager(
                verbose_name="Tags",
                through="pinboard.TaggedBookmark",
                to="pinboard.BookmarkTag",
                help_text="A comma-separated list of tags.",
            ),
        ),
    ]



from django.core.management.base import CommandError

from ditto.core.management.commands import DittoBaseCommand
from ditto.twitter.models import Account


class FetchTwitterCommand(DittoBaseCommand):
    # What we're fetching:
    singular_noun = "Tweet"
    plural_noun = "Tweets"

    # Child classes should supply some help text for the --recent argument:
    recent_help = ""

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            "--recent", action="store", default=False, help=self.recent_help
        )
        parser.add_argument(
            "--account",
            action="store",
            default=False,
            help="Only fetch for one Twitter account.",
        )

    def handle(self, *args, **options):
        # We might be fetching for a specific account or all (None).
        account = options["account"] if options["account"] else None

        if options["recent"]:
            # Will be either 'new' or a number; make the number an int.
            if options["recent"].isdigit():
                options["recent"] = int(options["recent"])

            results = self.fetch_tweets(account, options["recent"])
            self.output_results(results, options.get("verbosity", 1))
        elif options["account"]:
            msg = "Specify --recent as well as --account."
            raise CommandError(msg)
        else:
            msg = "Specify --recent, eg --recent=100 or --recent=new."
            raise CommandError(msg)

    def fetch_tweets(self, screen_name, count):
        """Child classes should override this method to call a method that
        fetches tweets and returns results, eg:
            return RecentTweetsFetcher(screen_name=screen_name).fetch(count=count)
        """
        return {}


class UpdateTwitterCommand(DittoBaseCommand):
    """Parent class for commands which update Tweets/Users already in the DB.

    Child classes probably only need a fetch() method.
    """

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            "--account",
            action="store",
            default=False,
            help="Which Twitter account to use for the API call.",
        )

    def handle(self, *args, **options):
        if options["account"]:
            screen_name = options["account"]
            try:
                Account.objects.get(user__screen_name=screen_name)
            except Account.DoesNotExist as err:
                msg = "There's no Account with a screen name of '{screen_name}'"
                raise CommandError(msg) from err
        else:
            msg = "Specify --account, eg --account=philgyford."
            raise CommandError(msg)

        results = self.fetch(screen_name)
        self.output_results(results, options.get("verbosity", 1))

    def fetch(self, screen_name):
        """Child classes should override this method to call a method that
        fetches Tweets/Users and returns results, eg:
            return TweetsFetcher(screen_name=screen_name).fetch()
        """
        return {}



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pinboard", "0008_bookmark_tags"),
    ]

    operations = [
        migrations.AlterField(
            model_name="bookmark",
            name="post_time",
            field=models.DateTimeField(
                help_text=b"The time this was created on Pinboard.",
                null=True,
                blank=True,
            ),
        ),
    ]



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("twitter", "0019_auto_20151028_1556"),
    ]

    operations = [
        migrations.RenameField(
            model_name="photo",
            old_name="media_url",
            new_name="url",
        ),
        migrations.RemoveField(
            model_name="photo",
            name="page_url",
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-23 14:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("pinboard", "0025_auto_20201223_1509"),
    ]

    operations = [
        migrations.AlterField(
            model_name="taggedbookmark",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_tagged_items",
                to="contenttypes.contenttype",
                verbose_name="content type",
            ),
        ),
        migrations.AlterField(
            model_name="taggedbookmark",
            name="tag",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="%(app_label)s_%(class)s_items",
                to="pinboard.bookmarktag",
            ),
        ),
    ]



import datetime

import factory
from taggit import models as taggit_models

from ditto.core.utils import datetime_now

from . import models


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.User

    nsid = factory.Sequence(lambda n: f"{n * 10000}@N01")
    username = factory.Sequence(lambda n: f"user{n:04}")
    realname = factory.Sequence(lambda n: f"User Name {n:04}")
    iconserver = 1234
    iconfarm = 5
    timezone_id = "America/Los_Angeles"
    photos_url = factory.Sequence(
        lambda n: f"https://www.flickr.com/photos/user{n:04}/"
    )

    photos_first_date = factory.LazyAttribute(
        lambda o: datetime_now() - datetime.timedelta(weeks=52)
    )
    photos_first_date_taken = factory.LazyAttribute(
        lambda o: datetime_now() - datetime.timedelta(weeks=52)
    )
    fetch_time = datetime_now()

    avatar = factory.django.ImageField(filename="my_avatar.jpg")


class AccountFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Account

    user = factory.SubFactory(
        UserFactory,
        username=factory.Sequence(lambda n: n),
        realname=factory.Sequence(lambda n: f"User Name {n:04}"),
    )


class PhotoFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Photo

    user = factory.SubFactory(
        UserFactory,
        username=factory.Sequence(lambda n: n),
        realname=factory.Sequence(lambda n: f"User Name {n:04}"),
    )

    flickr_id = factory.Sequence(lambda n: (n * 1000000))
    title = factory.Sequence(lambda n: f"Photo {n}")
    secret = factory.Sequence(lambda n: (n * 10000))
    original_secret = factory.Sequence(lambda n: (n * 10001))
    original_format = "jpg"
    server = "987"
    farm = 2
    media = "photo"
    post_time = datetime_now()
    taken_time = factory.LazyAttribute(
        lambda o: datetime_now() - datetime.timedelta(weeks=3)
    )
    last_update_time = factory.LazyAttribute(
        lambda o: datetime_now() - datetime.timedelta(weeks=2)
    )
    fetch_time = factory.LazyAttribute(
        lambda o: datetime_now() - datetime.timedelta(weeks=1)
    )

    original_file = factory.django.ImageField(filename="example.jpg")
    video_original_file = factory.django.FileField(filename="example.mov")

    thumbnail_width = 100
    thumbnail_height = 67
    small_width = 240
    small_height = 160
    small_320_width = 320
    small_320_height = 213
    medium_width = 500
    medium_height = 333
    medium_640_width = 640
    medium_640_height = 427
    medium_800_width = 800
    medium_800_height = 533
    large_width = 1024
    large_height = 683
    large_1600_width = 1600
    large_1600_height = 1067
    large_2048_width = 2048
    large_2048_height = 1365
    x_large_3k_width = 3072
    x_large_3k_height = 2048
    x_large_4k_width = 4096
    x_large_4k_height = 2731
    x_large_5k_width = 5120
    x_large_5k_height = 3413
    x_large_6k_width = 6000
    x_large_6k_height = 4000
    original_width = 6000
    original_height = 4000


class TagFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = taggit_models.Tag

    slug = factory.Sequence(lambda n: f"slug{n}")
    name = slug


class TaggedPhotoFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.TaggedPhoto

    tag = factory.SubFactory(TagFactory)
    flickr_id = factory.Sequence(lambda n: (n * 1000))
    author = factory.SubFactory(UserFactory)
    machine_tag = False
    content_object = factory.SubFactory(PhotoFactory)


class PhotosetFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Photoset

    user = factory.SubFactory(
        UserFactory,
        username=factory.Sequence(lambda n: n),
        realname=factory.Sequence(lambda n: f"User Name {n:04}"),
    )
    flickr_id = factory.Sequence(lambda n: (n * 1000000))
    title = factory.Sequence(lambda n: f"Photoset {n}")
    flickr_created_time = datetime_now()
    fetch_time = datetime_now()



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("twitter", "0021_video"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="photo",
            name="tweet",
        ),
        migrations.RemoveField(
            model_name="video",
            name="tweet",
        ),
        migrations.RemoveField(
            model_name="tweet",
            name="photos_count",
        ),
        migrations.AddField(
            model_name="tweet",
            name="media_count",
            field=models.PositiveSmallIntegerField(
                default=0,
                blank=True,
                help_text="Number of Photos/Videos attached to this Tweet",
            ),
        ),
        migrations.DeleteModel(
            name="Photo",
        ),
        migrations.DeleteModel(
            name="Video",
        ),
    ]



# Generated by Django 2.0.4 on 2018-04-16 11:44

from django.db import migrations


def re_save_tweets(apps, schema_editor):
    """
    Re-save all of the Tweets so that the HTML version of their text is updated
    using the newer version of Twython.
    """
    Tweet = apps.get_model("twitter", "Tweet")

    for tweet in Tweet.objects.all():
        tweet.save(update_fields=["text_html"])


class Migration(migrations.Migration):

    dependencies = [
        ("twitter", "0054_auto_20171113_1001"),
    ]

    operations = [
        migrations.RunPython(re_save_tweets),
    ]



from django.urls import path, re_path

from . import views

app_name = "flickr"

urlpatterns = [
    path("", view=views.HomeView.as_view(), name="home"),
    path("albums/", view=views.PhotosetListView.as_view(), name="photoset_list"),
    path("tags/", view=views.TagListView.as_view(), name="tag_list"),
    re_path(
        r"^tags/(?P<slug>[^/]+)/$",
        view=views.TagDetailView.as_view(),
        name="tag_detail",
    ),
    re_path(
        r"^(?P<nsid>[0-9N@]+)/$",
        view=views.UserDetailView.as_view(),
        name="user_detail",
    ),
    re_path(
        r"^(?P<nsid>[0-9N@]+)/albums/$",
        view=views.UserPhotosetListView.as_view(),
        name="user_photoset_list",
    ),
    re_path(
        r"^(?P<nsid>[0-9N@]+)/albums/(?P<flickr_id>[^/]+)/$",
        view=views.PhotosetDetailView.as_view(),
        name="photoset_detail",
    ),
    re_path(
        r"^(?P<nsid>[0-9N@]+)/tags/(?P<tag_slug>[^/]+)/$",
        view=views.UserTagDetailView.as_view(),
        name="user_tag_detail",
    ),
    re_path(
        r"^(?P<nsid>[0-9N@]+)/(?P<flickr_id>[0-9]+)/$",
        view=views.PhotoDetailView.as_view(),
        name="photo_detail",
    ),
]



# Generated by Django 1.9.5 on 2016-05-12 16:09

from django.db import migrations, models

import ditto.flickr.models


class Migration(migrations.Migration):

    dependencies = [
        ("flickr", "0013_photoset_photos_raw"),
    ]

    operations = [
        migrations.AddField(
            model_name="photo",
            name="original_file",
            field=models.FileField(
                blank=True, null=True, upload_to=ditto.flickr.models.Photo.upload_path
            ),
        ),
    ]



from datetime import datetime

from django.db import models


class WithScrobbleCountsManager(models.Manager):
    """
    Adds a with_scrobble_counts() method.
    """

    # Can we filter these (things) by Album?
    is_filterable_by_album = True

    # Can we filter these (things) by Artist?
    is_filterable_by_artist = True

    # Can we filter these (things) by Track?
    is_filterable_by_track = True

    def with_scrobble_counts(self, **kwargs):
        """
        Adds a `scrobble_count` field to the Queryset's objects, and
        orders the results by that, descending.

        eg:
            # All Tracks, each with a scrobble_count:
            Track.objects.with_scrobble_counts()

            # All Albums, each with a total scrobble_count:
            Album.objects.with_scrobble_counts()

            # All Artists, each with a total scrobble_count:
            Artist.objects.with_scrobble_counts()

            # Tracks by artist_obj:
            Track.objects.with_scrobble_counts(artist=artist_obj)

            # Tracks appearing on album_obj:
            Track.objects.with_scrobble_counts(album=album_obj)

            # Albums on which track_obj appears:
            Album.objects.with_scrobble_counts(track=track_obj)

        or combine filters:

            # Tracks by artist_obj, scrobbled by account_obj between
            # datetime_obj_1 and datetime_obj2:
            Track.objects.with_scrobble_counts(
                account         = account_obj,
                artist          = artist_obj,
                min_post_time   = datetime_obj_1,
                max_post_time   = datetime_obj_2,
            )

        Include an `account` to only include Scrobbles by that Account.
        Include an `album` to only include Scrobbles on that Album.
        Include an `artist` to only include Scrobbles by that Artist.
        Include a `track` to only include Scrobbles including that Track.
        Include a `min_post_time` to only include Scrobbles after then.
        Include a `max_post_time` to only include Scrobbles before then.
        """
        account = kwargs.get("account")
        min_post_time = kwargs.get("min_post_time")
        max_post_time = kwargs.get("max_post_time")
        album = kwargs.get("album")
        artist = kwargs.get("artist")
        track = kwargs.get("track")

        if album and not self.is_filterable_by_album:
            msg = "This is not filterable by album"
            raise ValueError(msg)

        if artist and not self.is_filterable_by_artist:
            msg = "This is not filterable by artist"
            raise ValueError(msg)

        if track and not self.is_filterable_by_track:
            msg = "This is not filterable by track"
            raise ValueError(msg)

        if account is not None and account.__class__.__name__ != "Account":
            msg = f"account must be an Account instance, not a {type(account)}"
            raise TypeError(msg)

        if album is not None and album.__class__.__name__ != "Album":
            msg = f"album must be an Album instance, not a {type(album)}"
            raise TypeError(msg)

        if artist is not None and artist.__class__.__name__ != "Artist":
            msg = f"artist must be an Artist instance, not a {type(account)}"
            raise TypeError(msg)

        if min_post_time is not None and type(min_post_time) is not datetime:
            msg = (
                "min_post_time must be a datetime.datetime, "
                f"not a {type(min_post_time)}"
            )
            raise TypeError(msg)

        if max_post_time is not None and type(max_post_time) is not datetime:
            msg = (
                "max_post_time must be a datetime.datetime, "
                f"not a {type(max_post_time)}"
            )
            raise TypeError(msg)

        filter_kwargs = {}

        if account:
            filter_kwargs["scrobbles__account"] = account

        if album:
            filter_kwargs["scrobbles__album"] = album

        if artist:
            filter_kwargs["scrobbles__artist"] = artist

        if track:
            filter_kwargs["scrobbles__track"] = track

        if min_post_time and max_post_time:
            filter_kwargs["scrobbles__post_time__gte"] = min_post_time
            filter_kwargs["scrobbles__post_time__lte"] = max_post_time

        elif min_post_time:
            filter_kwargs["scrobbles__post_time__gte"] = min_post_time

        elif max_post_time:
            filter_kwargs["scrobbles__post_time__lte"] = max_post_time

        qs = self.filter(**filter_kwargs)

        return qs.annotate(
            scrobble_count=models.Count("scrobbles", distinct=True)
        ).order_by("-scrobble_count")


class TracksManager(WithScrobbleCountsManager):
    """
    Adds a `scrobble_count` field to the Track objects.
    See WithScrobbleCountsManager for docs.
    """

    # We can't filter a list of Tracks by Tracks.
    is_filterable_by_track = False

    def with_scrobble_counts(self, **kwargs):
        "Pre-fetch all the Tracks' Artists."
        qs = super().with_scrobble_counts(**kwargs).prefetch_related("artist")
        return qs


class AlbumsManager(WithScrobbleCountsManager):
    """
    Adds a `scrobble_count` field to the Album objects.
    See WithScrobbleCountsManager for docs.
    """

    # We can't filter a list of Albums by Album.
    is_filterable_by_album = False

    def with_scrobble_counts(self, **kwargs):
        "Pre-fetch all the Albums' Artists."
        qs = super().with_scrobble_counts(**kwargs).prefetch_related("artist")
        return qs


class ArtistsManager(WithScrobbleCountsManager):
    """
    Adds a `scrobble_count` field to the Artist objects.
    See WithScrobbleCountsManager for docs.
    """

    # We can't filter a list of Artists by Artist.
    is_filterable_by_artist = False



import json
import os
from urllib.parse import urlparse

from django.core.files import File

from ditto.core.utils import datetime_now

from .fetch.savers import TweetSaver
from .models import Media


class IngestError(Exception):
    pass


class TweetIngester:
    """For importing a downloaded archive of tweets.
    Request yours from https://twitter.com/settings/account

    Use like:

    results = TweetIngester.ingest(
        '/Users/phil/Downloads/12552_dbeb4be9b8ff5f76d7d486c005cc21c9faa61f66/data/js/tweets'
    )

    Where that's the path to the directory containing the *.js files holding
    tweet data.

    results will be a dict of data about what happened, including
    results['success'] which is boolean.
    """

    def __init__(self):
        # Used as the 'fetch_time' for each tweet.
        self.fetch_time = datetime_now()

        # How many .js files we loaded the tweets from:
        self.file_count = 0

        # How mnay tweets we found in all the files:
        self.tweet_count = 0

        # How many media files we imported:
        self.media_count = 0

        # Stores all the imported data from the files before saving.
        # So that we know we've got through all the files within JSON errors
        # etc before we begin touching the DB.
        self.tweets_data = []

    def ingest(self, directory):
        """Import all the tweet data and create/update the tweets."""

        self._load_data(directory)

        self._save_tweets(directory)

        self._save_media(directory)

        if self.tweet_count > 0:
            return {
                "success": True,
                "tweets": self.tweet_count,
                "files": self.file_count,
                "media": self.media_count,
            }
        else:
            return {
                "success": False,
                "tweets": 0,
                "files": self.file_count,
                "media": self.media_count,
                "messages": ["No tweets were found"],
            }

    def _load_data(self, directory):
        """
        Child classes must implement their own _load_data() method.

        It should populate self.tweets_data with a list of dicts, each
        one being data about a single tweet.

        And it should set self.file_count to be the number of JS files
        we import the data from.
        """
        msg = (
            "Child classes of TweetImporter must implement their own "
            "_load_data() method."
        )
        raise NotImplementedError(msg)

    def _save_tweets(self, directory):
        """Go through the list of dicts that is self.tweets_data and
        create/update each tweet in the DB.
        """
        if len(self.tweets_data) == 0:
            return

        for tweet in self.tweets_data:
            TweetSaver().save_tweet(tweet, self.fetch_time)
            self.tweet_count += 1

    def _save_media(self, directory):
        """Save media files.
        Not doing anything by default.
        """


class Version1TweetIngester(TweetIngester):
    """
    Used for the old (original?) format of twitter archives, which contained
    three files and five folders, including data/js/tweets/ which contained
    a .js file for every month, like 2016_02.js. This is what we import
    the tweet data from.

    Sometime in 2019, between January and May, the format changed to
    what we call version 2.
    """

    def _load_data(self, directory):
        """Goes through all the *.js files in `directory` and puts the tweet
        data inside into self.tweets_data.

        No data is saved to the database until we've successfully loaded JSON
        from all of the files.

        Keyword arguments:
        directory -- The directory to load the files from.

        Raises:
        IngestError -- If the directory is invalid, or there are no .js files,
            or we can't load JSON from one of the files.
        """
        try:
            for file in os.listdir(directory):
                if file.endswith(".js"):
                    filepath = f"{directory}/{file}"
                    self._get_data_from_file(filepath)
                    self.file_count += 1
        except OSError as err:
            raise IngestError(err) from err

        if self.file_count == 0:
            msg = f"No .js files found in {directory}"
            raise IngestError(msg)

    def _get_data_from_file(self, filepath):
        """Looks in a file, parses its JSON, and adds a dict of data about
        each tweet found to self.tweets_data.

        Arguments:
        filespath -- Absolute path to the file.
        """
        with open(filepath) as f:
            lines = f.readlines()
            # Remove first line, that contains JavaScript:
            lines = lines[1:]
            try:
                tweets_data = json.loads("".join(lines))
            except ValueError as err:
                msg = f"Could not load JSON from {filepath}: {err}"
                raise IngestError(msg) from err
            else:
                self.tweets_data.extend(tweets_data)


class Version2TweetIngester(TweetIngester):
    """
    Used for what we call the version 2 format of twitter archive, which
    was introduced sometime between January and May of 2019.

    It contains two directories - assets and data - and a "Your archive.html" file.

    This not only saves the Tweet objects but also imports media from the
    tweet_media directory, saving it as Media files .
    """

    def __init__(self):
        super().__init__()

        # Tweets in this format don't contain data about the user.
        # So we create and store that once, in this separate field.
        self.user_data = {}

    def _load_data(self, directory):
        """
        Generate the user data and load all the tweets' data
        In this archive format, the tweets contain no data about the user.
        So we create a user data dict to use when saving the tweets.
        """

        self.user_data = self._construct_user_data(directory)

        self.tweets_data = self._get_json_from_file(directory, "tweet.js")

        self.file_count = 1

    def _save_tweets(self, directory):
        """
        Save the tweets with our constructed user data.
        """
        if len(self.tweets_data) == 0:
            return

        for tweet in self.tweets_data:
            # Here we pass in our user_data too, so save_tweet() can use
            # that in lieu of the data that is usually within each tweet's data.
            TweetSaver().save_tweet(tweet["tweet"], self.fetch_time, self.user_data)
            self.tweet_count += 1

    def _save_media(self, directory):
        """
        Save any animated gif's mp4 or an image's file for the saved tweets.
        """

        for t in self.tweets_data:
            tweet = t["tweet"]

            if "extended_entities" in tweet and "media" in tweet["extended_entities"]:
                for item in tweet["extended_entities"]["media"]:
                    try:
                        media_obj = Media.objects.get(twitter_id=int(item["id"]))
                    except Media.DoesNotExist:
                        pass
                    else:
                        if (
                            media_obj.media_type != "video"
                            and media_obj.has_file is False
                        ):
                            # We don't save video files - only image files, and mp4s
                            # for # GIFs - and only want to do this if we don't already
                            # have a file.

                            if (
                                media_obj.media_type == "animated_gif"
                                and media_obj.mp4_url
                            ):
                                url = media_obj.mp4_url
                            elif (
                                media_obj.media_type == "photo" and media_obj.image_url
                            ):
                                url = media_obj.image_url

                            if url:
                                # Work out name of file in the tweet_media directory:
                                parsed_url = urlparse(url)
                                filename = os.path.basename(parsed_url.path)
                                local_filename = f"{tweet['id_str']}-{filename}"
                                filepath = os.path.join(
                                    directory, "tweet_media", local_filename
                                )

                                with open(filepath, "rb") as f:
                                    django_file = File(f)
                                    if media_obj.media_type == "animated_gif":
                                        # When we fetch GIFs we also fetch an image file
                                        # for them. But their images aren't included in
                                        # the downloaded archive so we'll make do
                                        # without here.
                                        media_obj.mp4_file.save(filename, django_file)
                                        self.media_count += 1
                                    elif media_obj.media_type == "photo":
                                        media_obj.image_file.save(filename, django_file)
                                        self.media_count += 1

    def _construct_user_data(self, directory):
        """
        Make a single dict of data about a user like we'd get from the API.
        This data is in several separate files in the download so we need to
        piece it together from those.
        """

        account_data = self._get_json_from_file(directory, "account.js")

        profile_data = self._get_json_from_file(directory, "profile.js")

        verified_data = self._get_json_from_file(directory, "verified.js")

        try:
            user_data = {
                "id": int(account_data[0]["account"]["accountId"]),
                "id_str": account_data[0]["account"]["accountId"],
                "screen_name": account_data[0]["account"]["username"],
                "name": account_data[0]["account"]["accountDisplayName"],
                "profile_image_url_https": profile_data[0]["profile"]["avatarMediaUrl"],
                "verified": verified_data[0]["verified"]["verified"],
                # So that we don't mistake this for coming from the API when
                # we save the JSON:
                "ditto_note": (
                    "This user data was compiled from separate parts of a "
                    "downloaded twitter archive by Django Ditto"
                ),
            }
        except KeyError as err:
            msg = f"Error creating user data: {err}"
            raise ImportError(msg) from err

        return user_data

    def _get_json_from_file(self, directory, filepath):
        filepath = os.path.join(directory, filepath)
        try:
            f = open(filepath)  # noqa: SIM115
        except OSError as err:
            raise ImportError(err) from err

        lines = f.readlines()
        # Remove first line, that contains JavaScript:
        lines = ["["] + lines[1:]

        try:
            data = json.loads("".join(lines))
        except ValueError as err:
            msg = f"Could not load JSON from {filepath}: {err}"
            raise IngestError(msg) from err
        else:
            f.close()
            return data



from django.urls import path, re_path

from . import views

app_name = "pinboard"

urlpatterns = [
    path("", view=views.HomeView.as_view(), name="home"),
    path("to-read/", view=views.ToreadListView.as_view(), name="toread"),
    path("tags/", view=views.TagListView.as_view(), name="tag_list"),
    re_path(
        r"^tags/(?P<slug>[^/]+)/$",
        view=views.TagDetailView.as_view(),
        name="tag_detail",
    ),
    re_path(
        r"^(?P<username>\w+)/$",
        view=views.AccountDetailView.as_view(),
        name="account_detail",
    ),
    re_path(
        r"^(?P<username>\w+)/to-read/$",
        view=views.AccountToreadView.as_view(),
        name="account_toread",
    ),
    re_path(
        r"^(?P<username>\w+)/tags/(?P<tag_slug>[^/]+)/$",
        view=views.AccountTagDetailView.as_view(),
        name="account_tag_detail",
    ),
    re_path(
        r"^(?P<username>\w+)/(?P<hash>\w+)/$",
        view=views.BookmarkDetailView.as_view(),
        name="bookmark_detail",
    ),
]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("twitter", "0022_auto_20151104_0840"),
    ]

    operations = [
        migrations.CreateModel(
            name="Media",
            fields=[
                (
                    "id",
                    models.AutoField(
                        primary_key=True,
                        auto_created=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "time_created",
                    models.DateTimeField(
                        help_text="The time this item was created in the database.",
                        auto_now_add=True,
                    ),
                ),
                (
                    "time_modified",
                    models.DateTimeField(
                        help_text="The time this item was last saved to the database.",
                        auto_now=True,
                    ),
                ),
                (
                    "media_type",
                    models.CharField(
                        choices=[("photo", "Photo"), ("video", "Video")], max_length=8
                    ),
                ),
                ("twitter_id", models.BigIntegerField(unique=True)),
                ("image_url", models.URLField(help_text="URL of the image itself")),
                (
                    "is_private",
                    models.BooleanField(
                        help_text="If true, this item will not be shown on public-facing pages.",  # noqa: E501
                        default=False,
                    ),
                ),
                (
                    "large_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Large width", blank=True
                    ),
                ),
                (
                    "large_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Large height", blank=True
                    ),
                ),
                (
                    "medium_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Medium width", blank=True
                    ),
                ),
                (
                    "medium_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Medium height", blank=True
                    ),
                ),
                (
                    "small_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Small width", blank=True
                    ),
                ),
                (
                    "small_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Small height", blank=True
                    ),
                ),
                (
                    "thumb_w",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Thumbnail width", blank=True
                    ),
                ),
                (
                    "thumb_h",
                    models.PositiveSmallIntegerField(
                        null=True, verbose_name="Thumbnail height", blank=True
                    ),
                ),
                (
                    "mp4_url_1",
                    models.URLField(null=True, verbose_name="MP4 URL (1)", blank=True),
                ),
                (
                    "mp4_url_2",
                    models.URLField(null=True, verbose_name="MP4 URL (2)", blank=True),
                ),
                (
                    "mp4_url_3",
                    models.URLField(null=True, verbose_name="MP4 URL (3)", blank=True),
                ),
                (
                    "mp4_bitrate_1",
                    models.PositiveIntegerField(
                        null=True, verbose_name="MP4 Bitrate (1)", blank=True
                    ),
                ),
                (
                    "mp4_bitrate_2",
                    models.PositiveIntegerField(
                        null=True, verbose_name="MP4 Bitrate (2)", blank=True
                    ),
                ),
                (
                    "mp4_bitrate_3",
                    models.PositiveIntegerField(
                        null=True, verbose_name="MP4 Bitrate (3)", blank=True
                    ),
                ),
                (
                    "webm_url",
                    models.URLField(null=True, verbose_name="WebM URL", blank=True),
                ),
                (
                    "aspect_ratio",
                    models.CharField(help_text='eg, "4:3" or "16:9"', max_length=5),
                ),
                (
                    "duration",
                    models.PositiveIntegerField(
                        help_text="In milliseconds", null=True, blank=True
                    ),
                ),
                (
                    "tweet",
                    models.ForeignKey(to="twitter.Tweet", on_delete=models.CASCADE),
                ),
            ],
            options={"ordering": ["time_created"]},
        ),
    ]
